'''
module evaluation

Provides:

* Monte carlo classes and function to test of R&S procedures in 
simulated environments environments

'''

import numpy as np
import pandas as pd
from sklearn.model_selection import ParameterGrid

class AgentSimulation(object):
    '''
    Encapsulates a Monte-Carlo simulation framework for agents in 
    multi-arm bandit environment

    Agents must implement the interface

    solve()
    reset()
    '''

    def __init__(self, environment, agent, replications=1000):
        self._env = environment
        self._agent = agent
        self._reps = replications

    def simulate(self):

        best_indexes = np.zeros(self._reps, np.int32)

        for rep in range(self._reps):
            self._agent.reset()
            best_indexes[rep] = self._agent.solve()
            #best_indexes[rep] = self._agent.best_arm

        return best_indexes

class ExperimentResults(object):
    '''
    Results Container for an Agent Experiment
    '''
    def __init__(self, selections, correct_selections, p_correct_selections, 
                 opportunity_cost):
        self.selections = selections
        self.correct_selections = correct_selections
        self.p_correct_selections = p_correct_selections
        self.expected_opportunity_cost = opportunity_cost

class Experiment(object):
    '''
    Test the power of a given configuration of an agent
    at correct selection of a max of min 
    '''
    def __init__(self, env, procedure, best_index=0, objective='max', replications=1000):
        self._env = env
        self._agent = procedure
        self._sim = AgentSimulation(env, agent=procedure, replications=replications)
        self._best_index = best_index
        self._objective = objective
        self._reps = replications
    
    def execute(self):
        '''
        Execute the experiment
        '''
        selections = self._sim.simulate()
        
        correct_selections = (selections == self._best_index).sum()
        
        p_correct_selections = correct_selections / self._reps

        opportunity_cost = self._calculate_exp_opportunity_cost(selections)

        return ExperimentResults(selections, 
                                correct_selections, 
                                p_correct_selections, 
                                opportunity_cost)


    def _calculate_exp_opportunity_cost(self, selections):
        best_mean = self._env[self._env.best_design]._mu
        
        oc = 0.0
        for index in selections:
            oc += best_mean - self._env[index]._mu            
        return oc / len(selections)
        
    

class GridExperiment(object):

    def __init__(self, agent, environment, param_grid, 
                 best_index=0, replications=1000):
        self._agent = agent
        self._env = environment
        self._param_grid = ParameterGrid(param_grid)
        self._replications = replications
        self._best_index = best_index

    def fit(self):
        
        #create data frame to store results
        num_rows = len(self._param_grid)
        columns = list(self._param_grid[0].keys())
        columns.append('correct_selections')
        columns.append('p_correct_selections')
        
        df_results = pd.DataFrame(index = np.arange(0, num_rows),
                                  columns=columns)
       
        #simulate each of the agent configurations
        for index in range(len(self._param_grid)):
            for key in self._param_grid[index].keys():
                #set the agents attribute
               setattr(self._agent, key, self._param_grid[index][key])

               df_results.loc[index][key] = self._param_grid[index][key]

           
            experiment = Experiment(self._env, 
                                    self._agent, 
                                    best_index=self._best_index,
                                    replications=self._replications)

            results = experiment.execute()

            df_results.loc[index]['correct_selections'] = results.correct_selections
            df_results.loc[index]['p_correct_selections'] = results.p_correct_selections
            
        return df_results
                

            


                
                










'''
Toy models for testing optimisation via simulation procedures and
for running models.
'''

import numpy as np

class ManualOptimiser(object):
    '''
    A class to manually run individual and multiple replications
    from competiting simulated designs of a system.
    '''
    def __init__(self, model, n_designs, verbose=False):
        '''
        Constructor
        
        Parameters:
        -------
        model - object, simulation model that implements simulate(design_index)
        and register_observer(observer) methods
        
        n_designs - int, number of competiting designs 
        
        verbose_replications - bool, True if each individual rep observation
        is displayed.  False if run silently.  (default=False)
        '''
        self._model = model
        model.register_observer(self)
        self.verbose = verbose
        self._n_designs = n_designs
        self.means = np.zeros(n_designs, dtype=np.float)
        self.vars = np.zeros(n_designs, dtype=np.float)
        self._sq = np.zeros(n_designs, dtype=np.float)
        self._ses = np.zeros(n_designs, dtype=np.float)
        self.allocations = np.zeros(n_designs, np.int32)
        
    def __str__(self):
        return f"ManualOptimiser(model={self._model.__str__()}, n_designs={self._n_designs}, verbose={self.verbose})"    
    
    def simulate_designs(self, design_indexes=None, replications=1):
        '''
        Multiple replications of a simulated design
        
        Parameters:
        --------
        design_indexes - list, zero based design indexes.  If None
        then all system designs are included. (default=None)
        
        replications - int, number of replications to run (default=1)
        '''
        
        if design_indexes is None:
            design_indexes = [design for design in range(self._n_designs)]
        
        for design_index in design_indexes:
            self.simulate_design(design_index, replications)
    
    def simulate_design(self, design_index, replications=1):
        '''
        Multiple replications of a simulated design
        
        Parameters:
        --------
        design_index - int, zero based design index
        replications - int, number of replications to run (default=1)
        '''
        for rep in range(replications):
            self._model.simulate(design_index)
    
    def feedback(self, *args, **kwargs):
        '''
        Feedback from the simulation model
        Recieves a reward and updates understanding
        of an arm

        Keyword arguments:
        ------
        *args -- list of argument
                 0  sender object
                 1. design index
                 2. observation

        *kwards -- dict of keyword arguments:
                   None expected!

        '''
        design_index = args[1]
        observation = args[2]
        
        self.allocations[design_index] += 1
        self._update_moments(design_index, observation)
        
        if self.verbose:
            print(observation)
        
    def _update_moments(self, design_index, observation):
        '''
        Updates the running average, var of the design

        Parameters:
        ------
        design_index -- int, index of the array to update

        observation -- float, observation recieved from the last replication
        '''
        n = self.allocations[design_index]
        current_mean = self.means[design_index]
        new_mean = ((n - 1) / float(n)) * current_mean + (1 / float(n)) * observation

        if n > 1:
            self._sq[design_index] += (observation - abs(current_mean)) * (observation - abs(new_mean))
            self.vars[design_index] = self._sq[design_index] / (n - 1)
            self._ses[design_index] = np.sqrt(self.vars[design_index]) / np.sqrt(n)

        self.means[design_index] = new_mean
        


def guassian_sequence_model(start, end, step=1):
    bandits = [GaussianBandit(mean) for mean in range(start, end+1, step)]
    return BanditCasino(bandits)


def guassian_bandit_sequence(start, end, step=1):
    '''
    Sequence of GuassianBandit Arms.  Assumes unit variance. 
    e.g. start = 1, end = 10 return 10 bandits arms with means 1 through 10.

    Parameters:
    -------
    start - int, start of sequence of means (inclusive)
    end - int, end of sequence of means (exclusive)
    step - int, step size for sequence (default = 1)

    Returns:
    -------
    list of GaussianBandit objects ordered by means

    '''    
    return [GaussianBandit(mean) for mean in range(start, end, step)]
    

def random_guassian_model(mean_low, mean_high, var_low, var_high, n_designs):
    '''
    Create a model with n system designs where the mean and variance of the normal
    distributions are sampled to between the specified tolerances
    
    Parameters:
    -------
    mean_low - float, a lower bound on the means of the output distributions
    
    mean_high- float, an upper bound on the means
    
    var_low - float, a lower bound on the variance of the output distributions
    
    var_high - float, an upper bound on the variances.
    
    n_designs - int, the number of designs to create.
    
    Returns:
    --------
    BanditCasino with n_designs with means and variances between
    specified limits.
    '''
    means = np.random.uniform(low=mean_low, high=mean_high, size=n_designs)
    sigmas = np.random.uniform(low=var_low, high=var_high, size=n_designs)
    return custom_guassian_model(means, sigmas)



def custom_guassian_model(mus, sigmas):
    '''
    Creates a simulation model where each 
    output distribution is distributed N ~(mu, sigma)

    Assumes mus and signmas are of equal length

    Keyword arguments:
    ------
    mus - variable size list of means
    sigmas - list, variances

    Returns:
    ------
    object, simulation model 
    '''
    bandits = [GaussianBandit(mu, sigma) for mu, sigma in zip(mus, sigmas)]
    return BanditCasino(bandits)


class GaussianBandit(object):
    '''
    Classic one armed bandit gambling machine.

    A user plays the bandit by pulling its arm.

    The bandit returns a reward normally distribution 
    with mean mu and stdev sigma
    '''

    def __init__(self, mu, sigma=1.0):
        '''
        Constructor method for Gaussian Bandit

        Keyword arguments:
        -----
        mu -- float, mean of the normal distribution
        sigma -- float, stdev of the normal distribution (default = 1.0)

        '''
        self._mu = mu
        self._sigma = sigma
        self._number_of_plays = 0
        self._total_reward = 0
        self._observers = []

    def play(self):
        '''
        Pull the arm on the bernoulli bandit


        Returns:
        -----
        reward -- int with value = 0 when no reward
                  and 1 when the pull results in a win
        '''      
        reward = np.random.normal(self._mu, self._sigma)

        self._total_reward += reward
        self._number_of_plays += 1

        return reward

    def reset(self):
        '''
        Reset the number of plays and 
        total rewards to zero.
        '''
        self._number_of_plays = 0
        self._total_reward = 0




class BanditCasino(object):
    
    def __init__(self, bandits):
        '''
        Casino constructor method

        Keyword arguments:
        ------
        bandits -- list, of BernoulliBandits objects
        '''
        self._bandits = bandits
        self._current_index = 0
        self._observers = []

    def __str__(self):
        return f"BanditCasino()"
    
    def __getitem__(self, index):
        return self._bandits[index]

    def __get_number_of_arms(self):
        return len(self._bandits)

    def _get_best_arm(self):
        means = []
        for bandit in self._bandits:
            means.append(bandit._mu)
        
        return np.argmax(np.array(means))
        


    
    def simulate(self, bandit_index):
        '''
        Play a specific bandit machine.

        Notifies all observers of the outcome 

        Keyword arguments:
        -----
        bandit_index -- int, index of bandit to play 
        '''
        reward = self._bandits[bandit_index].play()
        self.notify_observers(bandit_index, reward)

    def random_action(self):
        '''
        Selects a bandit index at random and plays it.
        '''
        bandit_index = np.random.choice(len(self._bandits))
        self.simulate(bandit_index)

       
    def __iter__(self):
        return self

    def __next__(self):
        self._current_index += 1
        if self._current_index > len(self._bandits):
            self._current_index = 0
            raise StopIteration
        else:
            return self._bandits[self._current_index - 1]

    def register_observer(self, observer):
        self._observers.append(observer)
 
    def notify_observers(self, *args, **kwargs):
        for observer in self._observers:
            observer.feedback(self, *args, **kwargs) 

    number_of_arms = property(__get_number_of_arms)
    best_design = property(_get_best_arm)


'''
Contains algorithms for Optimisation via Simulation
with an indifference zone framework (fixed precision optimisation)

1. KN - (Kim-Nelson) sequential Ranking and Selection Algorithm

2. KNPlusPlus - Kim and Nelson updated their original KN procedure to 
update the variance estimate of designs at each stage.  

'''

import numpy as np

from .toy_models import (BanditCasino,    
                         GaussianBandit, 
                         guassian_bandit_sequence, 
                         guassian_sequence_model)

class KNPlusPlus(object):
    '''
    KN++ algorithm for Ranking and Selection

    More coordiation than KN. V
    Variances are updated at each stage.

    References.
    http://users.iems.northwestern.edu/~nelsonb/Publications/17KimNelson.pdf
    https://www2.isye.gatech.edu/~skim/KimNelson.pdf 

    '''
    def __init__(self, model, n_designs, delta, alpha=0.05, n_0=2, obj='max'):
        '''
        Constructor method for KN++  Ranking and Selection Procedure.

        Reference = 'TO ADD'

        Parameters:
        ----------

        model - object, simulation model that implements 
        method simulate(design, replications)

        n_designs - int, the number of designs (k)

        delta - float, the indifference zone

        alpha - float, between 0 and 1.  Used to calculate the 1-alpha 
        probability of correct selection

        n_0 - int, the number of initial observations (default = 2)
        
        obj - str, objective 'max' or 'min' (default='max)

        '''
        model.register_observer(self)
        self._env = model

        self._allocations = np.zeros(n_designs, np.int32)
        self._means = np.zeros(n_designs, np.float64)

        #sample variances for each design
        self._vars = np.zeros(n_designs, np.float64)
        #used for calculating running sample variance
        self._sq = np.zeros(n_designs, np.float64)

        #should I also be storing the variances of the differences?

        self._k = n_designs

        #set of non-eliminated systems
        #designs are screened at each stage and removed.
        self._contenders = np.arange(self._k)

        self._delta = delta
        self._alpha = alpha
        self._n_0 = n_0
        self._r = 0

        types = ['min', 'max']
        if obj not in types:
            raise ValueError('obj parameter must be min or max')

        #code set up for max negate for min.
        if obj == 'max':
            self._negate = 1.0
        else:
            self._negate = -1.0

        self._obj = obj


    def _calculate_eta(self):
        return 0.5 * (np.power(2 * (1 - np.power((1 - self._alpha), 1 / (self._k - 1))), -2/(self._n-1)) - 1)

    def __str__(self):
        return f"KN(n_designs={self._k}, delta={self._delta}, alpha={self._alpha}, n_0={self._n_0}, obj={self._obj})"

    def reset(self):
        '''Resets all attributes
        '''
        self._total_reward = 0
        self._allocations = np.zeros(self._k, np.int32)
        self._means = np.zeros(self._k, np.float64)
        self._allocations = np.zeros(self._k, np.int32)
        self._init_obs = np.zeros((self._k, self._n_0), np.float64)
        self._contenders = np.arange(self._k)
        self._vars = np.zeros(self._k, np.float64)
        self._sq = np.zeros(self._k, np.float64)
        self._n = 0


    def solve(self):
        '''Run KN++
        '''
        self.reset()
        self._initialisation()

        while not self._stopping():
            self._screening()
            self._sequential_replication()
            self._update()
            
        return self._contenders

    def _initialisation(self):  
        '''
        Initialise KN++
        
        Estimate initial sample means and variances of the differences

        '''      
        for observation in range(self._n_0):
            self._sequential_replication()

        self._eta = self._calculate_eta()
        self._h_squared = 2 * self._eta * (self._n - 1)

    def _stopping(self):
        '''
        If |I| = 1 then stop (and select the system whose index is in I)
        else take one further sample from each system

        '''
        if len(self._contenders) == 1:
            return True
        
        return False
                
    
    def _sequential_replication(self):
        '''
        Run a single replication of each
        design
        '''
        for design in self._contenders:
            self._env.simulate(design)
        
        self._n += 1

    def _screening(self):
        '''
        Loop through remaining contenders and screen if 
        mean(design_i) < mean(design_j) - epsilon_ij

        where epsilon_ij (w_ij in some papers) determines how far the sample mean 
        from system i can drop below the sample mean of system j without being eliminated.
        '''
        self._contenders_old = np.asarray(self._contenders).copy()
        contenders_mask = np.full(self._contenders.shape, True, dtype=bool)
                  
        #inefficient way to code this...will update to pure numpy at some stage

        for i in range(len(self._contenders_old)):
            for j in range(len(self._contenders_old)):

                if i != j:
                    design_i, design_j = self._contenders_old[i], self._contenders_old[j]
                    w_ij = self._elimination_distance(design_i, design_j)

                    if self._means[design_i] < self._means[design_j] - w_ij:
                        contenders_mask[i] = False
                        break
               
        self._contenders = self._contenders[contenders_mask]


    def _mean_difference(self, design_i, design_j):
        '''diff between two sample means
        '''
        mean_i = self._means[design_i]
        mean_j = self._means[design_j]
        return mean_i - mean_j
       

    def _elimination_distance(self, design_i, design_j):
        '''
        Returns the distance mean(design_i) can fall below mean(design_j)
        before it is ELIMINATED.

        Parameters:
        -----
        design_i - int, the index of the first design
        
        design_j - int, the index of the second design

        Returns:
        --------
        float. 
        '''
        #sample variances
        var_i = self._vars[design_i]
        var_j = self._vars[design_j]
        sum_of_vars = var_i + var_j

        #I think the problem is that I am not using the variance of the differences.
        w_ij = (self._delta / (2 * self._n)) \
            * (((self._h_squared * (sum_of_vars)) / self._delta**2) - self._n)

        #return max(0, w_ij)
        return w_ij


    def _update(self):
        '''update step.
        Recaulcate eta and h_squared
        '''
        self._eta = self._calculate_eta()
        self._h_squared = 2 * self._eta * (self._n - 1)


    def feedback(self, *args, **kwargs):
        '''
        Feedback from the simulated environment
        Recieves an observation from the system

        Parameters:
        ------
        *args -- list of argument
                 0  sender object
                 1. design index to update
                 2. observation

        *kwargs -- dict of keyword arguments

        '''
        design_index = args[1]
        observation = args[2]
        self._allocations[design_index] += 1 
                
        #update running mean and standard deviation of the design
        self._update_sample_estimates(design_index, observation)
        

    def _update_sample_estimates(self, design_index, observation):
        '''
        Updates the running sample mean and var of the design

        Parameters:
        ------
        design_index -- int, index of the array to update

        observation -- float, observation recieved from the last replication
        '''
        n = self._allocations[design_index]
        current_mean = self._means[design_index]
        new_mean = ((n - 1) / float(n)) * current_mean + (1 / float(n)) * observation * self._negate

        if n > 1:
            self._sq[design_index] += (observation - abs(current_mean)) * (observation - abs(new_mean))
            self._vars[design_index] = self._sq[design_index] / (n - 1)

        self._means[design_index] = new_mean    

    

    



class KN(object):
    def __init__(self, model, n_designs, delta, alpha=0.05, n_0=2, obj='max'):
        '''
        Constructor method for KN  Ranking and Selection Procedure.

        This works well for up to 20 competing designs

        References. 
        http://users.iems.northwestern.edu/~nelsonb/Publications/KimNelsonKN.pdf

        Parameters:
        ----------

        model - object, simulation model that implements 
        method simulate(design, replications)

        n_designs - int, the number of designs (k)

        delta - float, the indifference zone

        alpha - float, between 0 and 1.  Used to calculate the 1-alpha 
        probability of correct selection

        n_0 - int, the number of initial observations (default = 2)

        '''
        model.register_observer(self)
        self._env = model

        self._current_round = 0
        self._allocations = np.zeros(n_designs, np.int32)
        self._means = np.zeros(n_designs, np.float64)
        self._init_obs = np.zeros((n_designs, n_0), np.float64)
        
        self._k = n_designs

        #the set of non-eliminated systems I
        self._contenders = np.arange(self._k)

        self._delta = delta
        self._alpha = alpha
        self._n_0 = n_0
        
        #number of replications that has been run.
        self._r = 0

        self._eta = 0.5 * (np.power((2 * alpha) / (self._k - 1), -2/(self._n_0-1)) - 1)

        types = ['min', 'max']
        if obj not in types:
            raise ValueError('obj parameter must be min or max')

        #code set up for max negate for min.
        if obj == 'max':
            self._negate = 1.0
        else:
            self._negate = -1.0

        self._obj = obj


    def __str__(self):
        return f"KN(n_designs={self._k}, delta={self._delta}, alpha={self._alpha}, n_0={self._n_0}, obj={self._obj})"

    def reset(self):
        self._total_reward = 0
        self._allocations = np.zeros(self._k, np.int32)
        self._means = np.zeros(self._k, np.float64)
        self._allocations = np.zeros(self._k, np.int32)
        self._init_obs = np.zeros((self._k, self._n_0), np.float64)
        self._contenders = np.arange(self._k)
        self._r = 0
    
    def solve(self):
        '''Run procedure KN
        '''
        self.reset()
        self._initialisation()

        while not self._stopping():
            self._screening()
            self._sequential_replication()

        return self._contenders

    def _initialisation(self):  
        '''
        Initialise KN
        
        Estimate initial sample means and variances of the differences

        '''      
        self._h_squared = 2 * self._eta * (self._n_0 - 1)
        
        for observation in range(self._n_0):
            self._sequential_replication()
                
        self._variance_of_diffs = self._calc_variance_of_differences()

    def _sequential_replication(self):
        for design in self._contenders:
            self._env.simulate(design)
        
        self._r += 1

    def _calc_variance_of_differences(self):
        pairwise_diffs = self._init_obs[:, None] - self._init_obs
        variance_of_diffs = pairwise_diffs.var(axis=-1, ddof=1)
        #flattens array and drops differences with same design
        return variance_of_diffs[~np.eye(variance_of_diffs.shape[0],dtype=bool)]

    #need to check if this is correct...
    #possibly a bug.  Works when in designs are ordered, but not otherwise...
    #think this is fixed 28/02/2020 TM - to double check.
    def _screening(self):
        self._contenders_old = np.asarray(self._contenders).copy()
        contenders_mask = np.full(self._contenders.shape, True, dtype=bool)
        #terrible way to code this...!

        #designs in contention for this round
        for i in range(len(self._contenders_old)):
            for l in range(len(self._contenders_old)):
                if i != l:
                    design_i, design_l = self._contenders_old[i], self._contenders_old[l]

                    if not self._design_still_in_contention(design_i, design_l):
                        contenders_mask[i] = False
                        break
                

        self._contenders = self._contenders[contenders_mask]


    def _design_still_in_contention(self, design_i, design_l):
        w_il = self._limit_to_distance_from_sample_means(design_i, design_l)

        mean_i = self._means[design_i]
        mean_l = self._means[design_l]

        return mean_i >= mean_l - w_il

    def _limit_to_distance_from_sample_means(self, design_i, design_l):
        '''
        calculates W_li(r), 
        which determines how far the sample mean from system i can 
        drop below the sample means of system l without being eliminated
        '''
        #note - variance_of_diffs is a flat array for all comparisons and
        #requires this formaula to look up the value.
        index = design_i * (self._k - 1) + (design_l - 1)

        w_il = (self._delta / (2 * self._r)) \
            * (((self._h_squared * self._variance_of_diffs[index]) / self._delta**2) - self._r)
        
        return max(0, w_il)

    def _stopping(self):
        '''
        If |I| = 1 then stop (and select the system whose index is in I)
        else take one further sample from each system

        '''
        if len(self._contenders) == 1:
            return True
        
        return False

    def feedback(self, *args, **kwargs):
        '''
        Feedback from the environment
        Recieves a reward and updates understanding
        of an arm

        Keyword arguments:
        ------
        *args -- list of argument
                 0  sender object
                 1. design index to update
                 2. observation

        *kwards -- dict of keyword arguments:
                   None expected!

        '''
        design_index = args[1]
        observation = args[2]
        self._allocations[design_index] +=1 
        self._means[design_index] = self.updated_mean_estimate(design_index, observation)
        if self._r < self._n_0:
            self._init_obs[design_index][self._r] = observation
        
        
        
    def updated_mean_estimate(self, design_index, observation):
        '''
        Calculate the new running average of the design

        Keyword arguments:
        ------
        design_index -- int, index of the array to update
        observation -- float, observation recieved from the last action

        Returns:
        ------
        float, the new mean estimate for the selected arm
        '''
        n = self._allocations[design_index]
        current_value = self._means[design_index]
        new_value = ((n - 1) / float(n)) * current_value + (1 / float(n)) * observation * self._negate
        return new_value 

    









'''

Fixed budget alpirhtms

1. OCBA - optimal computing budget allocation
2. OCBA-m - optimal computering budget allocation top m.

'''

import numpy as np
import warnings


class OCBAM(object):
    '''
    Optimal Computer Budget Allocation Top M (OCBA-m)

    Given a total replication budget T allocate
    replications across designs in order to maximise the 
    approximate probability of correctly selecting the best m designs

    Assumes each system design has similar run time.

    Algorithm described in:
    -------------------------
    

    '''
    def __init__(self, model, n_designs, budget, delta, n_0=5, m=2, obj='min'):
        '''
        Constructor method for Optimal Budget Computer Allocation Top M

        Parameters:
        ---------

        model - object, simulation model that implements 
        interface action(design)

        n_designs - int, (k) the number of competing system designs 

        budget - int, (T) the total simulation budget available i.e. the 
        total number of replications available to allocate between systems

        delta - int, incremental budget to allocate.  Recommendation is
        > 5 and smaller than 10 percent of budget.  When simulation is expensive
        then this number could be set to 1.

        n_0 - int, the total number of initial replications.  Minimum allowed is 5
        (default=5)

        m - int, the best m designs

        obj - str, 'min' if minimisation; 'max' if maximisation.  (default='min')

        '''
        model.register_observer(self)
        self._env = model
        self._k = n_designs
        self._T  = budget
        self._delta = delta
        self._n_0 = n_0
        self._m = m
        self._obj = obj

        self._allocations = np.zeros(n_designs, np.int32)
        self._means = np.zeros(n_designs, np.float64)
        self._vars = np.zeros(n_designs, np.float64)
        self._ses = np.zeros(n_designs, np.float64)
        self._ratios = np.zeros(n_designs, np.float64)

        #used when calculating running standard deviation across designs
        #sum of squares
        self._sq = np.zeros(n_designs, np.float64)

        if self._obj == 'min':
            self._negate = -1.0
            self._min = True
        else:
            self._negate = 1.0
            self._min = False


    def __str__(self):
        return f'OCBA(n_designs={self._k}, m={self._m}, budget={self._T}, delta={self._delta}, n_0={self._n_0}, obj={self._obj})'


    def solve(self):
        '''
        This works okay for maximisation, but not if I 
        include negatation! What am I doing wrong!
        '''
        
        new_allocations = np.full(shape=self._k, 
                                  fill_value=self._n_0,
                                  dtype=np.int16)

        while self._allocations.sum() < self._T:
            
            #simulate systems using new allocation of budget
            self._simulate(new_allocations)
                        
            #calculate parameter c and deltas
            c = self._parameter_c(self._k, self._m)
            
            deltas = self._means - c

            #temp
            #deltas *= self._negate
            
            #allocate
            new_allocations = np.full(shape=self._k, 
                                  fill_value=0,
                                  dtype=np.int16)

            for i in range(self._delta):
                
                values = np.divide(self._allocations + new_allocations, np.square(np.divide(self._ses, deltas)))
                ranks = get_ranks(values)
                new_allocations[ranks.argmin()] += 1
        
        #return top m
        best = np.argpartition(self._means, -self._m)[-self._m:]


        return best

    def _simulate(self, new_allocations):
        '''
        For each design run n_0 initial replications
        '''
        for design in range(self._k):
            for replication in range(new_allocations[design]):
                self._env.simulate(design)

    def _parameter_c(self, k, m):
        order = np.argsort(self._means)
        s_ses = self._ses[order]
        s_means = self._means[order]

        return((s_ses[k-m+1] * s_means[k-m]) + (s_ses[k-m] * s_means[k-m+1]))/(s_ses[k-m]+s_ses[k-m+1]) 


    def feedback(self, *args, **kwargs):
        '''
        Feedback from the environment
        Recieves a reward and updates understanding
        of an arm

        Parameters:
        ------
        *args -- list of argument
                 0  sender object
                 1. arm index to update
                 2. observation

        *kwargs -- dict of keyword arguments

        '''
        design_index = args[1]
        observation = args[2]
        self._allocations[design_index] +=1 
                
        #update running mean and standard deviation
        self._update_moments(design_index, observation)
        

    def _update_moments(self, design_index, observation):
        '''
        Updates the running average, var of the design

        Parameters:
        ------
        design_index -- int, index of the array to update

        observation -- float, observation recieved from the last replication
        '''
        n = self._allocations[design_index]
        current_mean = self._means[design_index]
        new_mean = ((n - 1) / float(n)) * current_mean + (1 / float(n)) * observation * self._negate

        if n > 1:
            self._sq[design_index] += (observation - abs(current_mean)) * (observation - abs(new_mean))
            self._vars[design_index] = self._sq[design_index] / (n - 1)
            self._ses[design_index] = np.sqrt(self._vars[design_index]) / np.sqrt(n)

        self._means[design_index] = new_mean
        

def get_ranks(array):
    """
    Returns a numpy array containing ranks of numbers within a input numpy array
    e.g. [3, 2, 1] returns [2, 1, 0]
    e.g. [3, 2, 1, 4] return [2, 1, 0, 3]
        
    Parameters:
    --------
    array - np.ndarray, numpy array (only tested with 1d)

    Returns:
    ---------
    np.ndarray, ranks
        
    """
    temp = array.argsort()
    ranks = np.empty_like(temp)
    ranks[temp] = np.arange(len(array))
    return ranks


class OCBA(object):
    '''
    Optimal Computer Budget Allocation (OCBA)

    Given a total replication budget T allocate
    replications across designs in order to maximise the 
    probability of correct selections.

    Assumes each system design has similar run time.

    Algorithm described in:
    -------------------------
    Stochastic Simulation Optimization: An Optimal Computer Budget 
    Allocation.  by Chun-Hung Chen and Loo Hay Lee.

    Pages 49-50 for algorithm description
    Page 215 - for example C code.

    '''
    def __init__(self, model, n_designs, budget, delta, n_0=5, obj='min'):
        '''
        Constructor method for Optimal Budget Computer Allocation

        Parameters:
        ---------

        model - object, simulation model that implements 
        interface simulate(design)

        n_designs - int, (k) the number of competing system designs 

        budget - int, (T) the total simulation budget available i.e. the 
        total number of replications available to allocate between systems

        delta - int, incremental budget to allocate.  Recommendation is
        > 5 and smaller than 10 percent of budget.  When simulation is expensive
        then this number could be set to 1.

        n_0 - int, the total number of initial replications.  Minimum allowed is 5
        (default=5)

        obj - str, 'min' if minimisation; 'max' if maximisation.  (default='min')

        '''
        if n_0 < 5:
            raise ValueError('n_0 must be >= 5')

        if (budget - (n_designs * n_0)) % delta != 0:
            raise ValueError('(budget - (n_designs * n_0)) must be multiple of delta')

        types = ['min', 'max']
        if obj not in types:
            raise ValueError('obj parameter must be min or max')

        model.register_observer(self)
        self._env = model
        self._k = n_designs
        self._T  = budget
        self._delta = delta
        self._n_0 = n_0
        self._obj = obj

        self._allocations = np.zeros(n_designs, np.int32)
        self._means = np.zeros(n_designs, np.float64)
        self._vars = np.zeros(n_designs, np.float64)
        self._ratios = np.zeros(n_designs, np.float64)

        #used when calculating running standard deviation across designs
        #sum of squares
        self._sq = np.zeros(n_designs, np.float64)

        if obj == 'min':
            self._negate = 1.0
        else:
            self._negate = -1.0

    def __str__(self):
        return f'OCBA(n_designs={self._k}, budget={self._T}, delta={self._delta}, n_0={self._n_0}, obj={self._obj})'


    def reset(self):
        self._total_reward = 0
        self._current_round = 0
        self._allocations = np.zeros(self._k, np.int32)
        self._means = np.zeros(self._k, np.float64)
        self._vars = np.zeros(self._k, np.float64)
        self._ratios = np.zeros(self._k, np.float64)
        self._sq = np.zeros(self._k, np.float64)

    def solve(self):
        '''
        run the ranking and selection procedure
        Vanilla OCBA 
        '''
        self.reset()
        self._initialise()

        while self._allocations.sum() < self._T:
            new_allocations = self._allocate()
            
            for design in range(self._k):
                for replication in range(new_allocations[design]):
                    self._env.simulate(design)

        best = np.argmin(self._means)
        return best


    def _initialise(self):
        '''
        For each design run n_0 initial replications
        '''
        for design in range(self._k):
            for replication in range(self._n_0):
                self._env.simulate(design)

    
    def _allocate(self):
        '''
        Allocate the incremental budget across 
        designs
        '''

        #Notes could do with some cleaning and 
        #seperation into functions

        #total allocated + delta
        budget_to_allocate = self._allocations.sum() + self._delta

        #get indicies of best and second best designs so far
        #note treated as minimisation problem.  Means are negated if maximisation
        ranks = get_ranks(self._means) 
        best_index, s_best_index = np.argpartition(ranks, 2)[:2]

        self._ratios[s_best_index] = 1.0

        #Part 1: Ratio N_i / N_s
        #all 'select' does is exclude best and second best from arraywise calcs
        select = [i for i in range(self._k) if i not in [best_index, s_best_index]]

        temp = (self._means[best_index] - self._means[s_best_index]) \
            / (self._means[best_index] - self._means[select])

        self._ratios[select] = np.square(temp) * (self._vars[select] / self._vars[s_best_index])

        #Part 2: N_b
        #exclude best
        select = [i for i in range(self._k) if i not in [best_index]]
        temp = (np.square(self._ratios[select]) / self._vars[select]).sum()
        self._ratios[best_index] = np.sqrt(self._vars[best_index] * temp)

        #got all of the ratios now...
        more_runs = np.full(self._k, True, dtype=bool)
        additional_runs = np.zeros(self._k, dtype=np.float)
        more_alloc = True
        
        #do i need to use the more alloc or can i just use mask?
        while(more_alloc):
            more_alloc = False

            ratio_s = (more_runs * self._ratios).sum()

            additional_runs[more_runs] = \
                (budget_to_allocate / ratio_s) * self._ratios[more_runs]

            #additional_runs = additional_runs.astype(int)
            additional_runs = np.around(additional_runs).astype(int)
            
            mask = additional_runs < self._allocations
            additional_runs[mask] = self._allocations[mask]

            #disable designs where new allocation is less than has already been run.
            more_runs[mask] = 0  

            if mask.sum() > 0: 
                more_alloc = True

            if more_alloc:
                budget_to_allocate = self._allocations.sum() + self._delta
                budget_to_allocate -= (additional_runs * ~more_runs).sum()

        total_additional = additional_runs.sum()
        additional_runs[best_index] += self._allocations.sum() + self._delta - total_additional

        return additional_runs - self._allocations


    def feedback(self, *args, **kwargs):
        '''
        Feedback from the environment
        Recieves a reward and updates understanding
        of an arm

        Parameters:
        ------
        *args -- list of argument
                 0  sender object
                 1. arm index to update
                 2. observation

        *kwargs -- dict of keyword arguments

        '''
        design_index = args[1]
        observation = args[2]
        self._allocations[design_index] +=1 
                
        #update running mean and standard deviation
        self._update_moments(design_index, observation)
        

    def _update_moments(self, design_index, observation):
        '''
        Updates the running average, var of the design

        Parameters:
        ------
        design_index -- int, index of the array to update

        observation -- float, observation recieved from the last replication
        '''
        n = self._allocations[design_index]
        current_mean = self._means[design_index]
        new_mean = ((n - 1) / float(n)) * current_mean + (1 / float(n)) * observation * self._negate

        if n > 1:
            self._sq[design_index] += (observation - abs(current_mean)) * (observation - abs(new_mean))
            self._vars[design_index] = self._sq[design_index] / (n - 1)

        self._means[design_index] = new_mean



    


import setuptools
from sim_tools import __version__

# Read in the requirements.txt file
with open("requirements.txt") as f:
    requirements = []
    for library in f.read().splitlines():
        requirements.append(library)

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="sim-tools",
    #there must be an way to auto tick up the version number...
    version=__version__,
    author="Thomas Monks",
    author_email="t.m.w.monks@exeter.ac.uk",
    license="The MIT License (MIT)",
    description="Simulation Tools for Education and Practice",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/TomMonks/ovs-tutorial",
    packages=setuptools.find_packages(),
    #if true look in MANIFEST.in for data files to include
    include_package_data=False,
    #2nd approach to include data is include_package_data=False
    #package_data={"test_package": ["data/*.csv"]},
    #these are for documentation 
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6.9',
    install_requires=requirements,
)
